LIBRARY		libfli64
EXPORTS
	FLIOpen					@1
	FLIClose				@2
	FLISetDebugLevel		@3
	FLIGetLibVersion		@4
	FLIGetModel				@5
	FLIGetArrayArea			@6
	FLIGetVisibleArea		@7
	FLIExposeFrame			@8
	FLICancelExposure		@9
	FLIGetExposureStatus	@10
	FLISetTemperature		@11
	FLIGetTemperature		@12
	FLIGrabRow				@13
	FLIGrabFrame			@14
	FLIFlushRow				@15
	FLISetExposureTime		@16
	FLISetFrameType			@17
	FLISetImageArea			@18
	FLISetHBin				@19
	FLISetVBin				@20
	FLISetNFlushes			@21
	FLISetBitDepth			@22
	FLIReadIOPort			@23
	FLIWriteIOPort			@24
	FLIConfigureIOPort		@25
	FLIControlShutter		@26
	FLILockDevice			@27
	FLIUnlockDevice			@28
	FLIList					@29
	FLIFreeList				@30
	FLISetFilterPos			@31
	FLIGetFilterPos			@32
	FLIGetFilterCount		@33
	FLIStepMotor			@34
	FLIGetStepperPosition	@35
	FLIGetHWRevision		@36
	FLIGetPixelSize			@37
	FLIGetFWRevision		@38
	FLIHomeFocuser			@39
	FLICreateList			@40
	FLIDeleteList			@41
	FLIListFirst			@42
	FLIListNext				@43
	FLIControlBackgroundFlush @44
	FLISetDAC				@45
	FLIGetStepsRemaining	@46
	FLIStepMotorAsync		@47
	FLIReadTemperature		@48
	FLIGetFocuserExtent		@49
	FLIUsbBulkIO			@50
	FLIGetCoolerPower		@51
	FLIGetDeviceStatus		@52
	FLIGetCameraModeString	@53
	FLIGetCameraMode		@54
	FLISetCameraMode		@55
	FLIHomeDevice			@56
	FLIGrabVideoFrame		@57
	FLIStopVideoMode		@58
	FLIStartVideoMode		@59
	FLIGetSerialString		@60
	FLIEndExposure			@61
	FLITriggerExposure		@62
	FLISetFanSpeed			@63
	FLISetVerticalTableEntry @64
	FLIGetVerticalTableEntry @65
	FLIGetReadoutDimensions @66
	FLIEnableVerticalTable	@67
	FLIReadUserEEPROM		@68
	FLIWriteUserEEPROM		@69
	FLISetActiveWheel		@70
	FLIGetFilterName		@71
	FLIDebug				@72
